/*
 * A filesystem directory resource.
 *
 * Copyright (C) 2018, 2019 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#include <l4/sys/types.h>

#include <dirent.h>
#include <pthread.h>
#include <stdio.h>
#include <sys/stat.h>

#include <ext2fs/ext2fs.h>
#include <ipc/message.h>

/* Employed interface. */

#include "directory_listing_private_interface.h"



/* Utility macros. */

#define DIRENT_CORE_SIZE (sizeof(struct dirent) - sizeof(((struct dirent *) 0)->d_name))



/* Entry structure. */

struct entry
{
  /* Details of the entry provided during iteration. */

  struct ext2_dir_entry *dir_entry;

  /* Name and total entry sizes. The total entry size must be aligned for the
     architecture. This ensures that the serialised stat buffer is aligned. */

  int namelen;
  int reclen;

  /* Structure for population when providing metadata. */

  struct stat statbuf;

  /* Total size of the copied details. This must be aligned for the architecture
     for convenient access to the details as structures. */

  int size;
};



/* Filesystem directory resource employing an interface for private use by the
   server. */

class Fs_directory_listing : public DirectoryListingPrivate
{
private:
  ext2_filsys _fs;
  ext2_ino_t _dir;
  int _flags;

  /* Transfer buffer. */

  char *_buffer;
  size_t _buffer_size;

  /* Reading state. */

  size_t _seek_pos, _start_pos, _index;

  /* Concurrency and communication. */

  pthread_t _thread;
  l4_cap_idx_t _thread_ep;
  bool _terminating = false;

public:
  explicit Fs_directory_listing(ext2_filsys fs, ext2_ino_t dir, int flags=0)
  : _fs(fs), _dir(dir), _flags(flags) { }

  virtual ~Fs_directory_listing() { }

  /* Activation. */

  long start();

  /* Deallocation of resources. */

  void close();

  /* Accessor-operated methods. */

  size_t read_into(size_t start, char *buffer, size_t length);

  /* Thread mainloop. */

  void mainloop();

  /* Callback method for directory entry processing. */

  int record_entry(struct ext2_dir_entry *dir_entry);

  /* Entry handling for output. */

  void init_entry(struct entry *entry, struct ext2_dir_entry *dir_entry);

  /* Record writing methods. */

  int add_null_entry();
  void _copy_entry(char *out, struct entry *entry);
  void copy_entry(struct entry *entry);
  void copy_entry_start(struct entry *entry, int size);
  void copy_entry_end(struct entry *entry, int offset);

  /* Concurrency methods. */

  void _wait();
  void _yield(size_t size);

  /* IPC operation methods. */

  virtual long read(size_t position);
  virtual long terminate();

protected:
  /* Reply to requests for results. */

  void reply();
};
